#!/usr/bin/Rscript

## Name: RNAprof.R
## Version: 1.2.5
## Date: May 03, 2016
## Author: Van Du Tran and Daniel Gautheret
## Maintainer: Van Du Tran and Daniel Gautheret

##################################################
#### load libraries
##################################################
library(getopt)

##################################################
#### get options
spec <- matrix(c(
    'input', 'i', 1, "character", "library infomation file (required)",
    'gff', 'G', 1, "character", "input gff file (required)",
    'exon', 'e', 1, "character", "input exon file (required)",
    'figurepath', 'f', 2, "character", "figure path (optional, ./ by default)",
    'length', 'L', 2, "integer", "transcript length threshold (optional, 50 by default)",
    'minread', 'm', 2, "integer", "minimum mapped reads for a transcript (optional, 10 by default)",
    'candidate', 'C', 2, "integer", "minimum candidate length (optional, 1/2 length threshold by default)",
    'angle', 'a', 2, "double", "descent angle for dominant block normalization, [0,90], (optional, 89 by default)",
    'pval', 'P', 2, "double", "global p-value threshold for candidates (optional, 0.05 by default)",
    'pvalloc', 'p', 2, "double", "local p-value threshold for nucleotides (optional, 0.01 by default)",
    'epsilonread', 'E', 2, "integer", "maxium mapped read for lowly covered part to reduce (optional, 1/2 minread by default)",
    'coveredratio', 'c', 2, "double", "maximum covered ratio for lowly covered part to reduce, [0,1], (optional, 0.05 by default)",
    'reducedlocal', 'r', 2, "integer", "minimum length of a local reduced part (optional, 50 by default). A low-covered sub-profile longer than this will be reduced.",
    'reducedglobal', 'R', 2, "integer", "minimum length of a global reduced part (optional, 5000 by default). Region longer than this with several low-covered sub-profiles will be reduced.",
    'margin', 'M', 2, "integer", "margin to reduced part (< reduced/2) (optional, 10 by default)",
    'help', 'h', 0, "logical", "help"
    ), ncol=5, byrow=T)

##< check options
tryCatch( {
    opt = getopt(spec);
}, error=function(e) {
    cat(paste(getopt(spec, usage=T),"\n"));
    quit()
})

if (!is.null(opt$help) || is.null(opt$input) || is.null(opt$gff) || is.null(opt$exon)) {
    cat(paste(getopt(spec, usage=T),"\n"));
    quit();
}
##>

##< required arguments
inputFileName <- opt$input
geneFileName <- opt$gff
exonFileName <- opt$exon
##>

##< constants
MINREAD <- 10
EPSILONREAD <- MINREAD/2
TANDESCEND <- tan(89/180*pi)
PVTHRESHOLD <- 1e-3
LOCPVTHRESHOLD <- 1e-3
LOWCOVTHRESHOLD <- 0.05
MARGIN <- 10
MINREDUCEDSIZE <- 50
MAXREDUCEDSIZE <- 2000
LENTHRESHOLD <- 50
MINCANDSIZE <- ceiling(LENTHRESHOLD/2)
FIGUREPATH <- "./"
MINDOMINANTBLOCK <- 1/10
MAXDOMINANTBLOCK <- 1/4
INTRONWIDTH <- 1/10
EXONWIDTH <- 1/16
UTRWIDTH <- 1/12
PLOTMARGIN <- 16

##>

##-- gene length threshold
if (!is.null(opt$length)) {
    LENTHRESHOLD <- opt$length
}

##-- directory for output figures
if (!is.null(opt$figurepath)) {
    FIGUREPATH <- opt$figurepath
}

##-- minimum number of mapped reads
if (!is.null(opt$minread)) {
    MINREAD <- opt$minread
}

##-- minimum length of candidates
if (!is.null(opt$candidate)) {
    MINCANDSIZE <- opt$candidate
}

##-- descent angle for dominant block
if (!is.null(opt$angle)) {
    TANDESCEND <- tan(opt$angle/180*pi)
}

##-- p-value threshold for candidates
if (!is.null(opt$pval)) {
    PVTHRESHOLD <- opt$pval
}

##-- local p-value threshold for nucleotides
if (!is.null(opt$pvalloc)) {
    LOCPVTHRESHOLD <- opt$pvalloc
}

##-- maxium mapped read for lowly covered part to reduce
if (!is.null(opt$epsilonread)) {
    EPSILONREAD <- opt$epsilonread
}

##-- maximum covered ratio for lowly covered part to reduce
if (!is.null(opt$coveredratio)) {
    LOWCOVTHRESHOLD <- opt$coveredratio
}

##-- minimum length of reduced part 
if (!is.null(opt$reduced)) {
    MINREDUCEDSIZE <- opt$reduced
}

##-- margin to reduced part
if (!is.null(opt$margin)) {
    MARGIN <- opt$margin
}
##################################################

##################################################
#### load libraries
library(methods)
library(Cairo)
library(DESeq)
##################################################

##################################################
#### remove chromosome from the input profile
noChromosome <- function(v) {
    return (as.numeric(v[2:length(v)]))
}
##################################################

##################################################
#### max coverage on a no-chromosome profile
maxCoverage <- function(v) {
    return (max(v[-c(1,2)]))
}
##################################################

##################################################
#### geometric mean
gmean <- function(x) {
    len <- length(x)
    if (len==0) {
        return (0)
    }
    return (prod(x^(1/len)))
}
##################################################

##################################################
#### max quantile average
qtmax <- function(v) {
    sv <- sort(v, decreasing=T)
    i <- length(v)*MINDOMINANTBLOCK
    while (i < length(v)*MAXDOMINANTBLOCK && sv[i] < sv[i+1] + TANDESCEND) {
        i <- i + 1
    }
    
    return (mean(sv[1:(i-1)]))
}
##################################################

##################################################
#### calculate size factors for normalization
dominantBlockSizeFactors <- function(countData) {
    nr <- nrow(countData)
    nc <- ncol(countData)
    scd <- matrix(NA, nr, nc)
    
                                        #< calculate mean of countData
                                        #countMean <- apply(countData, 1, mean)
                                        #>
    
    ##< find size of high count data
    tmpCountData <- NULL
    maxlen <- LENTHRESHOLD

    for (j in 1:nc) {
        tmpCountData[[length(tmpCountData)+1]] <- countData[,j]
        maxCountData <- max(tmpCountData[[length(tmpCountData)]])
        tmpCountData[[length(tmpCountData)]] <- tmpCountData[[length(tmpCountData)]][tmpCountData[[length(tmpCountData)]] >= LOWCOVTHRESHOLD*maxCountData]
        maxlen <- max(maxlen, length(tmpCountData[[length(tmpCountData)]]))
    }
    ##>
    
    ##< sort count data
    for (j in 1:nc) {
        scd[,j] <- sort(countData[,j], decreasing=TRUE)
    }
    ##>
    
    ##< take only high sorted count data
    scd <- scd[1:maxlen,]
    ##>

    ##< find descending position
    i <- max(LENTHRESHOLD, maxlen*MINDOMINANTBLOCK)
    breakFlag <- FALSE
    while (i < maxlen*MAXDOMINANTBLOCK) {
        for (j in 1:nc) {
            if (scd[i,j] > scd[i+1,j] + TANDESCEND) {
                breakFlag <- TRUE
                break
            }
        }
        if (breakFlag) {
            break
        }
        i <- i + 1
    }
    ##>

    ##< calculate size factors
    scd <- scd[1:(i-1),]
    sf <- NULL
    sf <- apply(scd, 2, mean)
    
    ##< if a dominant block has less than EPSILONREAD reads
    if (length(sf[sf<=EPSILONREAD]) > 0) {
        return (NULL)
    }
    ##>
                                        #sf <- sf/mean(countMean[1:(i-1)])
    ##< normalize
    sf <- sf/gmean(sf)
    ##>
    
    return (sf)
}
##################################################

##################################################
#### calculate differential locations between two conditions
proDiff <- function(names, lim, LENGTH, curve1, curve2) {
    NUMREPLICATE1 <- nrow(curve1)
    NUMREPLICATE2 <- nrow(curve2)
    STEP <- 1
    localDiffs <- NULL
    local <- NULL
    localPval <- NULL
    localFC <- NULL
    localMean1 <- NULL
    localMean2 <- NULL
    localCurveMean1 <- array(list(), NUMREPLICATE1)
    localCurveMean2 <- array(list(), NUMREPLICATE2)
    
    libraries <- data.frame(row.names=names$label, condition=names$experiment)
    conditions <- libraries$condition
    conds <- unique(names$experiment)
    cond1 <- as.character(conds[1])
    cond2 <- as.character(conds[!conds==conds[1]])
    label1 <- as.character(names[names$experiment==cond1,]$label)
    label2 <- as.character(names[names$experiment==cond2,]$label)
                                        #print (libraries)
                                        #print (conditions)
    
    ##< count data for nbinomTest
    countData <- NULL
    for (i in seq(1, LENGTH-STEP+1, STEP)) {
        ii <- i+STEP-1
        nuclCov <- NULL
        for (j in 1:NUMREPLICATE1) {
            nuclCov <- c(nuclCov, max(ceiling(mean(curve1[j,i:ii])),1))
        }
        for (j in 1:NUMREPLICATE2) {
            nuclCov <- c(nuclCov, max(ceiling(mean(curve2[j,i:ii])),1))
        }
        countData <- rbind(countData, nuclCov)
    }
    rownames(countData) <- NULL
    ##>
                                        # countmean <- apply(countvect, 1, mean)
    
                                        # plot(countvect[,1], col=2, type="l")
                                        # points(countvect[,2], col=2, type="l")
                                        # points(countvect[,3], col=3, type="l")
                                        # points(countvect[,4], col=3, type="l")
                                        # points(countmean, col=1, type="l")
    
    rownames(countData) <- rownames(countData, do.NULL=FALSE, prefix="base")
    colnames(countData) <- rownames(libraries)
    
    locData <- newCountDataSet(countData, conditions)
                                        #print (head(counts(locData), 20))
                                        #locData <- estimateSizeFactors(locData)
    sizeft <- dominantBlockSizeFactors(countData)
    if (is.null(sizeft)) {
        return (NULL)
    }
    
    print ("Size factors:")
    print (sizeft)
    sizeFactors(locData) <- sizeft
    
                                        #sizeft <- sizeFactors(locData)
    
                                        #print (head(counts(locData, normalized=F), 20))
                                        #print (head(counts(locData, normalized=T), 20))
    normCounts <- counts(locData, normalized=T)
    
    locData <- try(estimateDispersions(locData, fitType="local", sharingMode="fit-only"), silent=TRUE)
                                        #, locfit_extra_args=list(maxk=500))#, lp_extra_agrs=list(nn=2)
    if (is(locData, "try-error")) {
        print ("TRY-ERROR: estimateDispersions on nucleotide")
        locData <- try(estimateDispersions(locData, fitType="local", sharingMode="fit-only", locfit_extra_args=list(maxk=300)), silent=TRUE)
    }
    if (is(locData, "try-error")) {
        print ("TRY-ERROR: estimateDispersions (locfit_extra_args) on nucleotide")
        return (NULL)
    }
    
                                        #quit()
                                        #print (head(counts(locData, normalized=T), 50))
                                        #print (tail(counts(locData, normalized=T), 50))
    nbtest <- nbinomTest(locData, cond1, cond2)
                                        #print (tail(nbtest,50))
                                        #quit()
                                        #quit()
    ##< search local candidates
    print ("Identifying candidates...")
    i <- 1
    while (i <= LENGTH) {
        while (i<=LENGTH &&
               (((is.na(nbtest$pval[i]) || is.infinite(nbtest$pval[i]) || nbtest$pval[i] > LOCPVTHRESHOLD) && !is.nan(nbtest$pval[i])) ||
               #(((is.na(nbtest$padj[i]) || is.infinite(nbtest$padj[i]) || nbtest$padj[i] > LOCPVTHRESHOLD) && !is.nan(nbtest$padj[i])) ||
                ##-- outliers
                sd(normCounts[i,label1]) >= min(
                      sd(c(min(normCounts[i,label2]), mean(normCounts[i,label1]))),
                      sd(c(max(normCounts[i,label2]), mean(normCounts[i,label1]))),
                      sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
                      sd(normCounts[i,c(label1,label2)])
                      ) ||
                sd(normCounts[i,label2]) >= min(
                      sd(c(min(normCounts[i,label1]), mean(normCounts[i,label2]))),
                      sd(c(max(normCounts[i,label1]), mean(normCounts[i,label2]))),
                      sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
                      sd(normCounts[i,c(label1,label2)])
                      )
                )
               ) {
            #print(c(i,nbtest$pval[i]))
            i <- i + 1
        }
        li <- i
        prevTrend <- (nbtest$foldChange[i] > 1)
        currentTrend <- (nbtest$foldChange[i] > 1)
                                        #avgscore <- 0
                                        #print (c("begin ",li))
        while (i<=LENGTH &&
               (((!is.na(nbtest$pval[i]) && is.finite(nbtest$pval[i]) && nbtest$pval[i] <= LOCPVTHRESHOLD) || is.nan(nbtest$pval[i])) &&
               #(((!is.na(nbtest$padj[i]) && is.finite(nbtest$padj[i]) && nbtest$padj[i] <= LOCPVTHRESHOLD) || is.nan(nbtest$padj[i])) &&
                prevTrend==currentTrend &&
                sd(normCounts[i,label1]) < min(
                    sd(c(min(normCounts[i,label2]), mean(normCounts[i,label1]))),
                    sd(c(max(normCounts[i,label2]), mean(normCounts[i,label1]))),
                    sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
                    sd(normCounts[i,c(label1,label2)])
                    ) &&
                sd(normCounts[i,label2]) < min(
                    sd(c(min(normCounts[i,label1]), mean(normCounts[i,label2]))),
                    sd(c(max(normCounts[i,label1]), mean(normCounts[i,label2]))),
                    sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
                    sd(normCounts[i,c(label1,label2)])
                    )
                )
               ) {
            ## print (normCounts[i,label1])
            ## print (normCounts[i,label2])
            ## print (c(sd(normCounts[i,label1]), sd(normCounts[i,label2]), sd(c(min(normCounts[i,label1]), mean(normCounts[i,label2]))),
            ##          sd(c(max(normCounts[i,label1]), mean(normCounts[i,label2]))),
            ##          sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
            ##          sd(normCounts[i,c(label1,label2)])))
            ## print (c(sd(normCounts[i,label1]), sd(c(min(normCounts[i,label2]), mean(normCounts[i,label1]))),
            ##                                      sd(c(max(normCounts[i,label2]), mean(normCounts[i,label1]))),
            ##                                      sd(c(mean(normCounts[i,label1]), mean(normCounts[i,label2]))),
            ##                                      sd(normCounts[i,c(label1,label2)])))
            #print(c("-------",i,nbtest$pval[i]))
            if (is.nan(nbtest$pval[i])) {
                nbtest$pval[i] <- 0
            }
            ## if (is.nan(nbtest$padj[i])) {
            ##     nbtest$padj[i] <- 0
            ## }
            i <- i + 1
            currentTrend <- (nbtest$foldChange[i] > 1)
        }
        ri <- i-1
                                        #print (c(li,ri))

        ##-- find local candidates
        if (li <= ri) {
            local <- c(local, li, ri)
            localPval <- c(localPval, mean(nbtest$pval[li:ri]))
            #localPval <- c(localPval, mean(nbtest$padj[li:ri]))
            localFC <- c(localFC, mean(nbtest$foldChange[li:ri]))
            localMean1 <- c(localMean1, mean(nbtest$baseMeanA[li:ri]))
            localMean2 <- c(localMean2, mean(nbtest$baseMeanB[li:ri]))
            for (k in 1:NUMREPLICATE1) {
                localCurveMean1[[k]] <- c(localCurveMean1[[k]], mean(curve1[k,li:ri]))
            }
            for (k in 1:NUMREPLICATE2) {
                localCurveMean2[[k]] <- c(localCurveMean2[[k]], mean(curve2[k,li:ri]))
            }
        }
    }
    
    ##< join local candidates
    print ("Joining candidates...")
                                        #print (local)
                                        #print (length(local))
                                        #print (head(nbtest))
                                        #print (localCurveMean1)
                                        #print (localCurveMean2)
    ## check trend of candidates
    ## for (i in seq(1,length(local),2)) {
    ##     print("*********************SUM")
    ##     print(local[i+1]-local[i]+1)
    ##     print(sum(nbtest$foldChange[(local[i]):(local[i+1])]>1))
    ##     print(sum(nbtest$foldChange[(local[i]):(local[i+1])]<1))
    ##     if (sum(nbtest$foldChange[(local[i]):(local[i+1])]>1)>0 && sum(nbtest$foldChange[(local[i]):(local[i+1])]<1)>0) {
    ##         print("--------------------------------")
    ##         print(local[i:(i+1)])
    ##     }
    ## }
    
    i <- length(local)-2
    while (i >= 2) {
        junc <- nbtest$foldChange[(local[i]+1):(local[i+1]-1)]
        trendjunc <- (junc > 1)
        trend1 <- (local[i] > 1 && nbtest$foldChange[local[i]] > 1)
        trend2 <- (local[i+1] < length(nbtest$foldChange) && nbtest$foldChange[local[i+1]] > 1)
        if (trend1==trend2 &&
            ((local[i+1]-local[i]-1 < MINREDUCEDSIZE &&
              max(mean(nbtest$baseMeanA[(local[i]+1):(local[i+1]-1)]), mean(nbtest$baseMeanB[(local[i]+1):(local[i+1]-1)])) <= max(EPSILONREAD,LOWCOVTHRESHOLD*max((nbtest$baseMeanA[(local[i-1]):(local[i])]), (nbtest$baseMeanB[(local[i+1]):(local[i+2])])))) ||
             (local[i+1]-local[i]-1 < min(max(local[i+2]-local[i+1]+1, local[i]-local[i-1]+1), max(MINCANDSIZE, 0.01*LENGTH)) &&
              length(trendjunc[trendjunc!=trend1])==0)
             )
            ) {
            locsizebefore <- (local[i]-local[i-1]+1)
            locsizeafter <- (local[i+2]-local[i+1]+1)
            locsizejoin <- locsizebefore + locsizeafter
            
            joinPval <- (localPval[(i/2)]*locsizebefore + localPval[(i/2)+1]*locsizeafter)/locsizejoin
            joinFC <- (localFC[(i/2)]*locsizebefore + localFC[(i/2)+1]*locsizeafter)/locsizejoin
            for (k in 1:NUMREPLICATE1) {
                joinCurveMean1 <- (localCurveMean1[[k]][(i/2)]*locsizebefore + localCurveMean1[[k]][(i/2)+1]*locsizeafter)/locsizejoin
                localCurveMean1[[k]] <- localCurveMean1[[k]][-(i/2)]
                localCurveMean1[[k]][(i/2)] <- joinCurveMean1
            }
            for (k in 1:NUMREPLICATE2) {
                joinCurveMean2 <- (localCurveMean2[[k]][(i/2)]*locsizebefore + localCurveMean2[[k]][(i/2)+1]*locsizeafter)/locsizejoin
                localCurveMean2[[k]] <- localCurveMean2[[k]][-(i/2)]
                localCurveMean2[[k]][(i/2)] <- joinCurveMean2
            }
            local <- local[c(-(i),-(i+1))]
            localPval <- localPval[-(i/2)]
            localPval[i/2] <- joinPval
            localFC <- localFC[-(i/2)]
            localFC[i/2] <- joinFC
        }
        i <- i-2
    }
    ##>
                                        ## print (local)
                                        ## print (localPval)
                                        ## print (localFC)
                                        ## print (length(local))

    ##< reduce a candidate to a nucleotide
    countDataCand <- NULL
    posCand <-NULL
    i <- 1
    j <- 1
    lenloc <- length(local)
    while (i <= lenloc) {
        ##< before candidate
        while (j < local[i]) {
            nuclCov <- NULL
            for (k in 1:NUMREPLICATE1) {
                nuclCov <- c(nuclCov, max(ceiling(curve1[k,j]),1))
            }
            for (k in 1:NUMREPLICATE2) {
                nuclCov <- c(nuclCov, max(ceiling(curve2[k,j]),1))
            }
            countDataCand <- rbind(countDataCand, nuclCov)
            j <- j+1
        }
        ##>

        ##< at candidate
        nuclCov <- NULL
        for (k in 1:NUMREPLICATE1) {
            nuclCov <- c(nuclCov, max(ceiling(localCurveMean1[[k]][(i+1)/2]),1))
        }
        for (k in 1:NUMREPLICATE2) {
            nuclCov <- c(nuclCov, max(ceiling(localCurveMean2[[k]][(i+1)/2]),1))
        }
        countDataCand <- rbind(countDataCand, nuclCov)
        posCand <- c(posCand, nrow(countDataCand))
        ##>
        
        j <- local[i+1] + 1
        i <- i+2
    }
    ##>
    
    ##< after the last candidate
    while (j <= LENGTH) {
        nuclCov <- NULL
        for (k in 1:NUMREPLICATE1) {
            nuclCov <- c(nuclCov, max(ceiling(curve1[k,j]),1))
        }
        for (k in 1:NUMREPLICATE2) {
            nuclCov <- c(nuclCov, max(ceiling(curve2[k,j]),1))
        }
        countDataCand <- rbind(countDataCand, nuclCov)
        j <- j+1
    }
    ##>
    
    rownames(countDataCand) <- NULL
    
                                        #print (local)
                                        #print (length(local))
                                        #print (posCand)
    rownames(countDataCand) <- rownames(countDataCand, do.NULL=FALSE, prefix="loc")
    colnames(countDataCand) <- rownames(libraries)
                                        #print ((countDataCand))
    locDataCand <- newCountDataSet(countDataCand, conditions)
                                        #print (tail(counts(locDataCand)))
                                        #locDataCand <- estimateSizeFactors(locDataCand)
    sizeFactors(locDataCand) <- sizeft
                                        #print (sizeFactors(locDataCand))
    locDataCand <- try(estimateDispersions(locDataCand, fitType="local", sharingMode="fit-only", locfit_extra_args=list(maxk=300)), silent=TRUE)

    lenloc <- length(local)
    ##< calculate fold change and p-value
    if (!is(locDataCand, "try-error")) {
        print ("TRY-ERROR: estimateDispersions on candidate")
        nbtestcand <- nbinomTest(locDataCand, cond1, cond2)
        if (lenloc > 1) {
            for (i in seq(1,lenloc,2)) {
                if (local[i+1]-local[i]+1 >= MINCANDSIZE) {
                    print ("Calculating p-value (without locfit error)...")
                    pvalue <- nbtestcand$pval[posCand[(i+1)/2]]
                    score <- nbtestcand$foldChange[posCand[(i+1)/2]]
                    meancand1 <- nbtestcand$baseMeanA[posCand[(i+1)/2]]
                    meancand2 <- nbtestcand$baseMeanB[posCand[(i+1)/2]]
                    if (is.na(pvalue)) {
                        pvalue <- 0
                    }
                    print (pvalue)
                    print (score)
                    localDiffs <- c(localDiffs, local[i], local[i+1], score, pvalue, meancand1, meancand2)
                }
            }
        }
        return (c(sizeft, localDiffs))
    }
    else {
        if (lenloc > 1) {
            for (i in seq(1,lenloc,2)) {
                if (local[i+1]-local[i]+1 >= MINCANDSIZE) {
                    print ("Calculating p-value (with locfit error)...")
                    pvalue <- localPval[(i+1)/2]
                    if (is.na(pvalue)) {
                        pvalue <- 0
                    }
                    score <- localFC[(i+1)/2]
                    meancand1 <- localMean1[(i+1)/2]
                    meancand1 <- localMean2[(i+1)/2]
                    print (pvalue)
                    print (score)
                    localDiffs <- c(localDiffs, local[i], local[i+1], score, pvalue, meancand1, meancand2)
                }
            }
        }
        return (c(sizeft, localDiffs))
    }
    ##>
}
##################################################

##################################################
#### plot curves
curvePlot <- function(LENGTH, reducedAnnotation, maxcov, sft, lim, curve1, curve2, cond1, cond2, rawcandidates, truecandidates, gene, geneid, rmblock, rmpch, rmstartpos, rmendpos, removelate, out) {
    if (out) {
        CairoPNG(file=paste(FIGUREPATH, geneid, ".png", sep=""), width=2000, height=500, res=75)
    }
  
    NUMREPLICATE1 <- nrow(curve1)
    NUMREPLICATE2 <- nrow(curve2)
    NUMREPLICATE <- NUMREPLICATE1 + NUMREPLICATE2

    ##< draw only candidates of p < PVTHRESHOLD
    candidates <- rawcandidates[1:NUMREPLICATE]
    for (j in seq(NUMREPLICATE+1,length(rawcandidates),6)) {
        if (as.numeric(rawcandidates[j+3]) < PVTHRESHOLD) {
            candidates <- c(candidates, rawcandidates[j:(j+3)])
        }
    }
    ##>
    
    ##< if there are candidates
    if (length(candidates) > NUMREPLICATE) {
        rawlen <- lim[2] - lim[1] + 1
        numDigits <- nchar(as.character(rawlen)) + 1
        ## idrep <- 1
        ## for (j in 1:NUMREPLICATE1) {
        ##     curve1[j,] <- curve1[j,]/candidates[idrep]
        ##     idrep <- idrep + 1
        ## }
        ## for (j in 1:NUMREPLICATE2) {
        ##     curve2[j,] <- curve2[j,]/candidates[idrep]
        ##     idrep <- idrep + 1
        ## }
        
        newcandidates <- NULL
        for (j in seq(NUMREPLICATE+1,length(candidates),4)) {
            newcandidates <- c(newcandidates, candidates[j] - lim[1] + 1, candidates[j+1] - lim[1] + 1)
        }
        
        ##< calculate necessary plot margin for printing densed removed blocks
        lenrmblock <- length(rmblock)
        i <- 1
        j <- i + 1
        maxDense <- 1
        OVERMARGINPLOT <- FALSE
        while (i <= lenrmblock && j <= lenrmblock) {
            if (rmblock[j]-rmblock[i] > LENGTH/PLOTMARGIN) {
                i <- i + 1
            }
            j <- j + 1
            maxDense <- max(maxDense, j - i)
        }
        if (maxDense > 8) {
            maxDense <- 1
            OVERMARGINPLOT <- TRUE
        }
        ##>

        ##< calculate necessary plot margin for printing densed candidates
        lennewcand <- length(newcandidates)
        i <- 1
        j <- i + 2
        maxDenseCand <- 1
        while (i <= lennewcand && j <= lennewcand) {
            if (0.5*(candidates[j*2+NUMREPLICATE-1]+candidates[j*2+NUMREPLICATE])-0.5*(candidates[i*2+NUMREPLICATE-1]+candidates[i*2+NUMREPLICATE]) > LENGTH/PLOTMARGIN) {
                i <- i + 2
            }
            j <- j + 2
            maxDenseCand <- max(maxDenseCand, (j - i)/2)
        }
        ##>
                                        #print (maxDenseCand)
                                        #print (maxDense)
        ##< plot
        par(oma=c(maxDense-1,0,maxDenseCand*2-1,0))
        plot(reducedAnnotation, col="black", type="S", xlab="", ylab="Coverage", ylim=c(-maxcov*EXONWIDTH*2,maxcov))
        points(-reducedAnnotation-maxcov*INTRONWIDTH*2, col="black", type="S")
        
        ##-- coverage profiles
        for (j in 1:NUMREPLICATE1) {
            greenIndex <- 64*(j-1) %% 255
            points(curve1[j,], col=rgb(255,greenIndex,0,maxColorValue=255), type="l")
                                        #print (curve1[j,])
        }
        for (j in 1:NUMREPLICATE2) {
            greenIndex <- 64*(j-1) %% 255
            points(curve2[j,], col=rgb(0,greenIndex,255,maxColorValue=255), type="l")
                                        #print (curve2[j,])
        }
        segments(0, -maxcov*INTRONWIDTH, 1, -maxcov*INTRONWIDTH)

        ##-- direction info
        if (gene[7]=="+") {
            text(-length(curve1[1,])/80, -maxcov*INTRONWIDTH, "5'", cex=3)
        }
        else {
            text(-length(curve1[1,])/80, -maxcov*INTRONWIDTH, "3'", cex=3)
        }
        
        i <- 1
        while (i <= lenrmblock) {
            ##-- reduced block plot
            points(rmblock[i], 1, t="p", col="black", pch=rmpch[i], bg="black", cex=3)
            if (i==1 || rmblock[i]-rmblock[i-1] > LENGTH/PLOTMARGIN) {
                linepos <- 2
            }
            else {
                linepos <- 2 + ((linepos-1)%%maxDense)
            }
            ##-- reduced block coordinates
            if (!OVERMARGINPLOT) {
                mtext(paste(".",(rmstartpos[i]+lim[1]-1)%%(10^numDigits),"-.",(rmendpos[i]+lim[1]-1)%%(10^numDigits),sep=""), line=linepos, at=rmblock[i], side=1, cex=1, font=3, outer=F)
            }
            i <- i + 1
        }

        jprev <- 0
        for (j in seq(1,length(newcandidates),2)) {
            if (as.numeric(candidates[j*2+NUMREPLICATE+2]) < PVTHRESHOLD) {
                ## && (as.numeric(candidates[j*2+NUMREPLICATE+1])>=1.5 || as.numeric(candidates[j*2+NUMREPLICATE+1])<=2/3)) {
                ##< candidate boundary
                abline(v=candidates[j*2+NUMREPLICATE-1]-length(removelate[removelate<candidates[j*2+NUMREPLICATE-1]]), col=rgb(127,0,255,maxColorValue=255))
                abline(v=candidates[j*2+NUMREPLICATE]-length(removelate[removelate<candidates[j*2+NUMREPLICATE]]), col=rgb(127,127,255,maxColorValue=255))
                segments(candidates[j*2+NUMREPLICATE-1]-length(removelate[removelate<candidates[j*2+NUMREPLICATE-1]]),
                         0.99*maxcov,
                         candidates[j*2+NUMREPLICATE]-length(removelate[removelate<candidates[j*2+NUMREPLICATE]]),
                         0.99*maxcov, col="green", lwd=2)
                
                ##>
                
                ##< candidate info
                shiftremovelateprev <- length(removelate[removelate<candidates[jprev*2+NUMREPLICATE-1]]) + length(removelate[removelate<candidates[jprev*2+NUMREPLICATE]])
                shiftremovelate <- length(removelate[removelate<candidates[j*2+NUMREPLICATE-1]]) + length(removelate[removelate<candidates[j*2+NUMREPLICATE]])
                if (jprev==0 || 0.5*(candidates[j*2+NUMREPLICATE-1]+candidates[j*2+NUMREPLICATE]-shiftremovelate)-0.5*(candidates[jprev*2+NUMREPLICATE-1]+candidates[jprev*2+NUMREPLICATE]-shiftremovelateprev) > LENGTH/PLOTMARGIN) {
                    linepos <- maxDenseCand*2
                    lineposscore <- 1
                }
                else {
                    linepos <- maxDenseCand*2 + ((linepos+1-maxDenseCand*2)%%(maxDenseCand*2))
                    lineposscore <- (lineposscore+2)%%(maxDenseCand*2)
                }
                mtext(paste(".",truecandidates[j]%%(10^numDigits),"-.",truecandidates[j+1]%%(10^numDigits),sep=""), line=linepos, side=3, at=0.5*(candidates[j*2+NUMREPLICATE-1]+candidates[j*2+NUMREPLICATE]-shiftremovelate), col=rgb(127,0,255,maxColorValue=255), font=3, outer=F)
                mtext(paste("S:", round(candidates[j*2+NUMREPLICATE+1],digits=2), sep=""), line=lineposscore, side=3, at=0.5*(candidates[j*2+NUMREPLICATE-1]+candidates[j*2+NUMREPLICATE]-shiftremovelate), col=rgb(127,0,255,maxColorValue=255), outer=F)
                mtext(paste("p:", format(candidates[j*2+NUMREPLICATE+2],digits=2,scientific=T), sep=""), line=lineposscore-1, side=3, at=0.5*(candidates[j*2+NUMREPLICATE-1]+candidates[j*2+NUMREPLICATE]-shiftremovelate), col=rgb(127,0,255,maxColorValue=255), outer=F)
                ##>
                
                jprev <- j
            }
        }
        ##-- gene info
        sft1 <- paste(round(sft[1:NUMREPLICATE1],digits=1),collapse=", ")
        sft2 <- paste(round(sft[(NUMREPLICATE1+1):NUMREPLICATE],digits=1),collapse=", ")
        mtext(paste(gene[1], ":", gene[4], "-", gene[5], " (", gene[7], " ", geneid, "), gene size = ", rawlen, " bp, expressed size = ", LENGTH, " bp, size factors = ", cond1, "(red): [", sft1 , "]-", cond2, "(blue): [", sft2 , "]", sep=""), line=3+maxDense, at=0, adj=0, side=1, cex=1.5, font=2, outer=F)
        ##>
    }
    ##>
    
    if (out) {
        invisible(dev.off())
    }
}
##################################################

##################################################
#### histogram on expressed percentage
histPlot <- function(vect, scale, cond1, cond2, suffix, type, out) {
    if (out) {
        CairoPNG(file=paste(FIGUREPATH, cond1, "_", cond2, "_", suffix, ".png", sep=""), width=500, height=500, res=75)
    }
                                        #d <- density(vect)
    
                                        #hist(vect, breaks=100, xlim=c(0,scale*100), ylim=c(0,max(d$y*scale*100)), xlab=type, ylab="N° transcripts", col="red")
    hist(vect, breaks=100, xlim=c(0,scale*100), xlab=type, ylab="N° transcripts", col="red", main=paste("Distribution of ", type, sep=""))
    
    ##points(d$x, d$y*scale*100, type="l", col="blue")
    
    if (out) {
        invisible(dev.off())
    }
}
##################################################

##################################################
#### expressed percentage vs length plot
expressedPlot <- function(len, expressedVect, cond1, cond2, xname, yname, out) {
    if (out) {
        CairoPNG(file=paste(FIGUREPATH, cond1, "_", cond2, "_Expressed.png", sep=""), width=500, height=500, res=75)
    }

    plot(len, expressedVect, xlab=xname, ylab=yname, ylim=c(0,100))
                                        #fit <- glm(expressedVect~len)
                                        #abline(fit)
    
    if (out) {
        invisible(dev.off())
    }
}
##################################################

##################################################
#### histogram on candidates
##################################################
candHistPlot <- function(res, cond1, cond2, out) {
    if (out) {
        CairoPNG(file=paste(FIGUREPATH, cond1, "_", cond2, "_HistPlot.png", sep=""), width=500, height=500, res=75)
    }
    
    logPVal <- -log10(as.numeric(res[,5]))
    if (length(logPVal[is.infinite(logPVal)==FALSE])>0) {
        logPVal[is.infinite(logPVal)==TRUE] <- max(logPVal[is.infinite(logPVal)==FALSE]) + 10
    }
    else {
        logPVal[is.infinite(logPVal)==TRUE] <- 20
    }
    
    hist(logPVal, breaks=100, xlab="p-value", ylab="N° candidates", col="red", main="Distribution of candidate p-values")
    
    if (out) {
        invisible(dev.off())
    }
}

##################################################
#### Volcano Plot
volcanoPlot <- function(res, hthreshold, vthreshold, cond1, cond2, out) {
    if (out) {
        CairoPNG(file=paste(FIGUREPATH, cond1, "_", cond2, "_VolcanoPlot.png", sep=""), width=500, height=500, res=75)
    }
    
    logHT <- -log10(as.numeric(hthreshold))
    logVT <- log2(as.numeric(vthreshold))
    logPVal <- -log10(as.numeric(res[,5]))
    logFC <- log2(as.numeric(res[,4]))
    
    if (length(logPVal[is.infinite(logPVal)==FALSE])>0) {
        logPVal[is.infinite(logPVal)==TRUE] <- max(logPVal[is.infinite(logPVal)==FALSE]) + 10
    }
    else {
        logPVal[is.infinite(logPVal)==TRUE] <- 20
    }
    
    if (length(logFC[is.infinite(logFC)==FALSE])>0) {
        logFC[is.infinite(logFC)==TRUE] <- min(logFC[is.infinite(logFC)==FALSE]) - 4
    }
    else {
        logFC[is.infinite(logFC)==TRUE] <- -8
    }
    
    if (length(logFC)>0) {
        xlim=c(-max(abs(logFC)),max(abs(logFC)))
        ylim=c(0,max(logPVal))
    }
    else {
        xlim <- c(-2,2)
        ylim <- c(0,10)
    }
    plot(logFC, logPVal, pch=20, col=ifelse((logFC > logVT | logFC < -logVT) & (logPVal > logHT), "red", "black"), xlim=xlim, ylim=ylim , main="Volcano Plot", xlab="log2FC", ylab="-log10pval")
    abline(h=logHT)
    
    if (out) {
        invisible(dev.off())
    }
}
##################################################

##################################################
#### load input data
loadFileName <- function(inputFile){
    return (read.table(inputFile, header=T, sep="\t"))
}
##################################################

##################################################
#### main function
main <- function(inputFileName, geneFileName, exonFileName) {
    ##< initialize result table
    restab <- NULL
    ##>
    
    ##< load file name for each condition
    names <- loadFileName(inputFileName)
    conds <- unique(names$experiment)
    cond1 <- as.character(conds[1])
    cond2 <- as.character(conds[!conds==conds[1]])
    condName1 <- as.character(names[names$experiment==cond1,]$file)
    condName2 <- as.character(names[names$experiment==cond2,]$file)
    NUMREPLICATE1 <- length(condName1)
    NUMREPLICATE2 <- length(condName2)
    NUMREPLICATE <- NUMREPLICATE1 + NUMREPLICATE2
                                        #print (condName1)
                                        #print (condName2)
    ##>
    
    ##< load library size for each library and calculate their mean size
    libSize1 <- as.numeric(names[names$experiment==cond1,]$size)
    libSize2 <- as.numeric(names[names$experiment==cond2,]$size)
    if (NUMREPLICATE==0) {
        print ("Libraries not found")
        quit()
    }
    rootExp <- 1/NUMREPLICATE
    meanSize <- prod(libSize1^rootExp)*prod(libSize2^rootExp)
    ##>
    
    ##< load coverage profiles
    ##-- coverage vector of whole gene
    vect1 <- vector("list", NUMREPLICATE1)
    vect2 <- vector("list", NUMREPLICATE2)
    ##-- coverage vector of well covered part of gene (at least EPSILONREAD reads for each library)
    vectcovered1 <- vector("list", NUMREPLICATE1)
    vectcovered2 <- vector("list", NUMREPLICATE2)
    ##-- connections for input files
    con1 <- vector("list", NUMREPLICATE1)
    con2 <- vector("list", NUMREPLICATE2)
    for (j in 1:NUMREPLICATE1) {
        con1[[j]] <- file(condName1[j], open="r")
    }
    for (j in 1:NUMREPLICATE2) {
        con2[[j]] <- file(condName2[j], open="r")
    }
    conexon <- file(exonFileName, open="r")
    congene <- file(geneFileName, open="r")
    ##>
    
    ##-- lengths of whole genes
    lengeneVect <- NULL
    MAXLENGENE <- 0
    ##-- lengths of well covered part of genes (at least EPSILONREAD in a libary)
    lencoveredgeneVect <- NULL
    MAXLENCOVEREDGENE <- 0
    ##-- vector of expressedPercent
    expressedVect <- NULL

    ##< remove result file if there exists
    resgff <- paste(FIGUREPATH, cond1,"_",cond2,".gff",sep="")
    if (!is.na(file.info(resgff)$size)) {
        file.remove(resgff)
    }
    ressft <- paste(FIGUREPATH, cond1,"_",cond2,".sft",sep="")
    if (!is.na(file.info(ressft)$size)) {
        file.remove(ressft)
    }
    cat(c("chr", "start", "end", "id", rep(cond1, NUMREPLICATE1), rep(cond2, NUMREPLICATE2), "\n"), file=ressft, append=T, sep="\t")
    ##>
    
    ##< run comparison for each gene
    ##-- number of genes
    NUMGENE <- 0
    ##-- number of genes with candidates
    NUMGENEFOUND <- 0
    while (length(line <- readLines(congene, n=1)) > 0) {
        NUMGENE <- NUMGENE + 1
 
        ##-- read gene
        genes <- (strsplit(line, "\t"))[[1]]
        ##-- gene attributes
        geneid <- strsplit(genes[9], "[ ;=\"]+")
        
        ##< read gene profiles 
        for (j in 1:NUMREPLICATE1) {
      	    vect1[[j]] <- (strsplit(readLines(con1[[j]], n=1), "\t"))[[1]]
        }
        for (j in 1:NUMREPLICATE2) {
            vect2[[j]] <- (strsplit(readLines(con2[[j]], n=1), "\t"))[[1]]
        }
        ##>
        
        print (c("TRANSCRIPT POSITION: ", paste(vect1[[1]][1], vect1[[1]][2], vect1[[1]][3], sep=" ")))
        
        ##< max coverage in each libary
        vect1 <- lapply(vect1, noChromosome)
        vect2 <- lapply(vect2, noChromosome)
        maxcov1 <- sapply(vect1, maxCoverage)
        maxcov2 <- sapply(vect2, maxCoverage)
        ##>
        
        ##-- gene position
        lim <- vect1[[1]][1:2]
        ##-- gene length
        lengene <- lim[2] - lim[1] + 1
        ##-- vector of gene lengths
        lengeneVect <- c(lengeneVect, lengene)
        
        ##< remove low-count segments
        ##-- starting position (of removed block) vector obtained after removal
        rmblock <- NULL
        ##-- pch type of rmblock
        rmpch <- NULL
        ##-- removed position vector
        remove <- NULL
        ##-- removed starting position vector
        rmstartpos <- NULL
        ##-- removed ending position vector
        rmendpos <- NULL
        ##-- total removed length
        rmlen <- 0
        k <- 3
        while (k <= lengene + 2) {
            nucltaken <- FALSE

            ##-- starting position of removed block
            rmstart <- k - 2
                        
            ##< advance until where the coverage is high enough in at least one replicate
            while (k <= lengene + 2) {
                for (j in 1:NUMREPLICATE1) {
                    if (vect1[[j]][k] > max(EPSILONREAD,LOWCOVTHRESHOLD*maxcov1[j])) {
                        nucltaken <- TRUE
                        break
                    }
                }
                for (j in 1:NUMREPLICATE2) {
                    if (vect2[[j]][k] > max(EPSILONREAD,LOWCOVTHRESHOLD*maxcov2[j])) {
                        nucltaken <- TRUE
                        break
                    }
                }
                k <- k + 1
                if (nucltaken) {
                    break
                }
            }
            ##>

            ##-- ending position of removed block
            rmend <- k - 2 - 2
            
            ##< if block size is larger than MINREDUCEDSIZE, then reduce
            if (rmend-rmstart >= MINREDUCEDSIZE) {
                rmstart <- rmstart + MARGIN
                rmend <- rmend - MARGIN
                if (rmend < rmstart) {
                    print ("Failed: MINREDUCEDSIZE < 2*MARGIN")
                    quit()
                }
                remove <- c(remove, rmstart:rmend)
                rmblock <- c(rmblock, rmstart-rmlen)
                rmpch <- c(rmpch, 6)
                rmstartpos <- c(rmstartpos, rmstart)
                rmendpos <- c(rmendpos, rmend)
                rmlen <- rmlen + rmend - rmstart + 1
            }
            ##>
        }
        ## print("PCH")
        ## print(j)
        ## print(rmpch)
        ## print(rmblock)
        ## print(rmstartpos+lim[1]-1)
        ## print(rmendpos+lim[1]-1)
        
        ##< early: reduced profile state for candidate identification
        rmstartposearly <- rmstartpos
        rmendposearly <- rmendpos
        ##>
        
        ##< apply removal to get covered gene profiles
        for (j in 1:NUMREPLICATE1) {
            vectcovered1[[j]] = vect1[[j]][-c(1,2,remove)]
        }
        for (j in 1:NUMREPLICATE2) {
            vectcovered2[[j]] = vect2[[j]][-c(1,2,remove)]
        }
        lencoveredgene <- length(vectcovered1[[1]])
        lencoveredgeneVect <- c(lencoveredgeneVect, lencoveredgene)
        ##>
        ##>

        ##< check coverage profile vector lengths
        curve1 <- NULL
        curve2 <- NULL
        for (j in 1:NUMREPLICATE1) {
            curve1 <- rbind(curve1, vectcovered1[[j]])
            if (lengene != length(vect1[[j]])-2) {
                print ("Unequal profile lengths")
                quit()
            }
        }
        for (j in 1:NUMREPLICATE2) {
            curve2 <- rbind(curve2, vectcovered2[[j]])
            if (lengene != length(vect2[[j]])-2) {
                print ("Unequal profile lengths")
                quit()
            }
        }
        ##>
        
        ##< normalize by library size
        mean1 <- NULL
        mean2 <- NULL
        curve1 <- curve1*(meanSize/libSize1)
        curve2 <- curve2*(meanSize/libSize2)
        mean1 <- apply(curve1, 1, mean)
        mean2 <- apply(curve2, 2, mean)
        maxcov <- max(curve1, curve2)
        ##>
        
        ##< expressed percentage
        expressedPercent <- lencoveredgene/lengene*100
        expressedVect <- c(expressedVect, expressedPercent)
        MAXLENGENE <- max(MAXLENGENE, lengene)
        MAXLENCOVEREDGENE <- max(MAXLENCOVEREDGENE, lencoveredgene)
                                        #print (expressedPercent)
        ##>
        
        ##< load exon list
        exons <- strsplit(readLines(conexon, n=1), "\t")
        ##>

        ##-- profile difference candidates
        res <- NULL
        sft <- NA
        ##-- candidate information to be written in loci.gff
        geneall <- NULL
        print ("Checking constraints on mapped reads...")
        ##-- only consider genes with at least 'MINREAD' reads in at least one condition, and 'LENTHRESHOLD' nucleotides
        if (lencoveredgene >= LENTHRESHOLD && maxcov >= MINREAD) {
            ##< calculate profile difference
            tryCatch(
                {
                    res <- proDiff(names, lim, lencoveredgene, curve1, curve2)
                    sft <- round(res[1:NUMREPLICATE]*c(libSize1,libSize2)/meanSize, digits=2)
                    cat(c(genes[1], genes[4:5], geneid[[1]][2], sft, "\n"), file=ressft, append=T, sep="\t")
                }, error = function(e) {
                    print (c("Warning: unable to perform DESeq on ", paste(genes[1], genes[4], genes[5], geneid[[1]][2], sep=" ")))
                    cat(c(genes[1], genes[4:5], geneid[[1]][2], rep("NA",NUMREPLICATE), "\n"), file=ressft, append=T, sep="\t")
                    return (0)
                }
                )
            ##>

            ##< check if there is candidate of p-value < PVTHRESHOLD
            pv <- NUMREPLICATE + 4
            curvePlotFlag <- FALSE
            while (pv <= length(res)) {
                if (res[pv] < PVTHRESHOLD) {
                    curvePlotFlag <- TRUE
                    break
                }
                pv <- pv + 6
            }
            ##>
            
            print ("Printing...")
            ##-- if there exists candidates
            if (length(res) > NUMREPLICATE) {
                NUMGENEFOUND <- NUMGENEFOUND + 1

                removelate <- NULL
                if (lencoveredgene > MAXREDUCEDSIZE) {
                    candstarts <- res[sort(c(seq(NUMREPLICATE+1,length(res),6), seq(NUMREPLICATE+2,length(res),6)))]
                    candstarts <- c(1,candstarts,lencoveredgene)
                    rmlen <- 0
                    
                    for (j in seq(2,length(candstarts),1)) {
                        rmblockloc <- rmblock[rmblock>candstarts[j-1] & rmblock<candstarts[j]]
                        rmblockloclen <- length(rmblockloc)
                        candinside <- (j%%2!=0)
                        ##-- if outside candidates
                        if (!candinside) {
                            if (rmblockloclen >= 4 || (rmblockloclen>=1 && rmblockloc[rmblockloclen]-rmblockloc[1] > MAXREDUCEDSIZE)) {
                                rmblockstart <- rmblockloc[1]
                                rmblockend <- rmblockloc[rmblockloclen]
                                removelate <- c(removelate, rmblockstart:rmblockend)
                                
                                rmstartpos <- c(rmstartpos[which(rmblock<candstarts[j-1])], rmstartpos[which(rmblock==rmblockstart)], rmstartpos[which(rmblock>candstarts[j])])
                                rmendpos <- c(rmendpos[which(rmblock<candstarts[j-1])], rmendpos[which(rmblock==rmblockend)], rmendpos[which(rmblock>candstarts[j])])
                                rmpch <- c(rmpch[which(rmblock<candstarts[j-1])], 5, rep(6,length(rmblock[rmblock>candstarts[j]])))
                                
                                rmblock <- c(rmblock[rmblock<candstarts[j-1]], rmblockstart-rmlen, rmblock[rmblock>candstarts[j]])
                                
                                rmlen <- rmlen + rmblockend - rmblockstart + 1
                                ## print("PCH")
                                ## print(j)
                                ## print(rmpch)
                                ## print(rmblock)
                                ## print(rmstartpos)
                                ## print(candstarts)
                            }
                            else {
                                rmblock[rmblock>candstarts[j-1] & rmblock<candstarts[j]] <- rmblock[rmblock>candstarts[j-1] & rmblock<candstarts[j]] - rmlen
                            }
                        }
                        ##-- if inside candidates
                        else {
                            rmblock[rmblock>candstarts[j-1] & rmblock<candstarts[j]] <- rmblock[rmblock>candstarts[j-1] & rmblock<candstarts[j]] - rmlen
                        }
                    }
                    if (length(removelate)>0) {
                        curve1 <- curve1[,-removelate]
                        curve2 <- curve2[,-removelate]
                    }
                }

                ##< recover true coordinates of candidates
                trueres <- NULL
                krm <- 1
                rmstartposearly <- c(rmstartposearly, lengene+1)
                rmendposearly <- c(rmendposearly, lengene+1)
                lenkept <- rmstartposearly[1] - 1
                lenrm <- 0
                for (j in seq(NUMREPLICATE+1,length(res),6)) {
                    while (krm < length(rmendposearly) && res[j] > lenkept) {
                        lenkept <- lenkept + rmstartposearly[krm+1] - rmendposearly[krm] - 1
                        lenrm <- lenrm + rmendposearly[krm] - rmstartposearly[krm] + 1
                        krm <- krm + 1
                    }
                    trueres <- c(trueres, res[j] + lenrm + lim[1] - 1)
                    while (krm < length(rmendposearly) && res[j+1] > lenkept) {
                        lenkept <- lenkept + rmstartposearly[krm+1] - rmendposearly[krm] - 1
                        lenrm <- lenrm + rmendposearly[krm] - rmstartposearly[krm] + 1
                        krm <- krm + 1
                    }
                    trueres <- c(trueres, res[j+1] + lenrm + lim[1] - 1)
                }
                rmstartposearly <- rmstartposearly[-length(rmstartposearly)]
                rmendposearly <- rmendposearly[-length(rmendposearly)]
                ##>
                                                
                ##< calculate annotation coornidates 
                exons <- as.numeric(exons[[1]])
                exons <- c(lim[1], exons[1], exons, exons[length(exons)], lim[2])
                exons <- exons - lim[1] + 1
                ##>

                ##-- max normalized coverage
                curve1 <- curve1/res[1:NUMREPLICATE1]
                curve2 <- curve2/res[(NUMREPLICATE1+1):NUMREPLICATE]
                maxcov <- 1.1*max(curve1,curve2)
                
                ##< coordinates of annotation for plot
                ##-- introns
                annotation <- rep(-maxcov*INTRONWIDTH, lengene+1)
                
                ##-- UTR or extension
                if (exons[1]+1 <= exons[2]) {
                    annotation[(exons[1]+1):exons[2]] <- -maxcov*UTRWIDTH
                }
                for (k in seq(3,length(exons)-2,2)) {
                    ##-- exons
                    if (exons[k]+1 <= exons[k+1]) {
                        annotation[(exons[k]+1):exons[k+1]] <- -maxcov*EXONWIDTH
                    }
                }
                ##-- UTR or extension
                if (exons[length(exons)-1]+1 <= exons[length(exons)]) {
                    annotation[(exons[length(exons)-1]+1):exons[length(exons)]] <- -maxcov*UTRWIDTH
                }
                ##>
                
                ##< take exons from highly covered parts
                if (length(remove) > 0) {
                    reducedAnnotationHigh <- annotation[-remove]
                }
                else {
                    reducedAnnotationHigh <- annotation
                }
                if (length(removelate) > 0) {
                    reducedAnnotation <- reducedAnnotationHigh[-removelate]
                }
                else {
                    reducedAnnotation <- reducedAnnotationHigh
                }
                ##>
                
                ##< calculate size of extended region
                extendedSize <- length(reducedAnnotationHigh[reducedAnnotationHigh==-maxcov*UTRWIDTH])
                ##>
               
                ##< plot candidates
                if (curvePlotFlag) {
                    curvePlot(lencoveredgene, reducedAnnotation, maxcov, sft, lim, curve1, curve2, cond1, cond2, res, trueres, genes[1:8], geneid[[1]][2], rmblock, rmpch, rmstartpos, rmendpos, removelate, TRUE)
                }
                ##>

                ##< write candidates into file
                geneid <- strsplit(genes[9], ";")
                for (j in seq(NUMREPLICATE+1,length(res),6)) {
                    restab <- rbind(restab, c(genes[1], genes[7], genes[9], res[j+2], res[j+3]))
                    geneall[[length(geneall)+1]] <- c(genes[1:3], format(trueres[(j-NUMREPLICATE+1)/2],scientific=F), format(trueres[(j-NUMREPLICATE+1)/2+1],scientific=F), genes[6:8], paste(geneid[[1]][1], ":prof", (j-NUMREPLICATE-1)/4+1,";", paste(geneid[[1]][-1], collapse=";"), ";Exp_extension=", round(extendedSize/lencoveredgene, digits=3), ";Size=", as.numeric(trueres[(j-NUMREPLICATE+1)/2+1])-as.numeric(trueres[(j-NUMREPLICATE+1)/2])+1, ";Score=", round(res[j+2],digits=2), ";Mean1=", round(res[j+4],digits=2), ";Mean2=", round(res[j+5],digits=2), ";P-val=", format(res[j+3],digits=2,scientific=T), "\n", sep=""))
                    cat((geneall[[length(geneall)]]), file=resgff, append=T, sep="\t")
                }
                ##>
            }
        }
        else {
            print ("Failed checking constraints on mapped reads...")
            cat(c(genes[1], genes[4:5], geneid[[1]][2], rep("NA",NUMREPLICATE), "\n"), file=ressft, append=T, sep="\t")
        }
    }

    close(conexon)
    close(congene)
    for (j in 1:NUMREPLICATE1) {
        close(con1[[j]])
    }
    for (j in 1:NUMREPLICATE2) {
        close(con2[[j]])
    }
    
    ##< histogram
    #histPlot(expressedVect, 1, cond1, cond2, "EFT", "Expressed fraction of transcript (%)", TRUE)
    #histPlot(lengeneVect, ceiling(MAXLENGENE/100), cond1, cond2, "WTS", "Transcript size", TRUE)
    #histPlot(lencoveredgeneVect, ceiling(MAXLENCOVEREDGENE/100), cond1, cond2, "CTS", "Covered transcript size", TRUE)
    #expressedPlot(lengeneVect, expressedVect, cond1, cond2, "Transcript size", "Expressed fraction (%)", TRUE)
    ##>
    
    if (is.null(restab)) {
        NUMEVENT <- 0
    }
    else {
        NUMEVENT <- nrow(restab)
    }
    
    if (NUMEVENT > 0) {
        ##< candidate histogram
        candHistPlot(restab, cond1, cond2, TRUE)
        ##>
        
        ##< Volcano Plot
        volcanoPlot(restab, 1e-5, 4, cond1, cond2, TRUE)
        ##>
    }
    
    report <- paste(FIGUREPATH, cond1,"_",cond2,"_Report.txt",sep="")
    if (!is.na(file.info(report)$size)) {
        file.remove(report)
    }
    for (j in 1:NUMREPLICATE1) {
        cat(paste("Number of reads ",cond1, j, ": ", libSize1[j], "\n", sep=""), file=report, append=T, sep="\t")
    }
    for (j in 1:NUMREPLICATE2) {
        cat(paste("Number of reads ",cond2, j, ": ", libSize2[j], "\n", sep=""), file=report, append=T, sep="\t")
    }
    cat(paste("Number of genes: ", NUMGENE, "\n", sep=""), file=report, append=T, sep="\t")
    cat(paste("Number of events: ", NUMEVENT, "\n", sep=""), file=report, append=T, sep="\t")
    cat(paste("Number of genes with events: ", NUMGENEFOUND, "\n", sep=""), file=report, append=T, sep="\t")
    
    return (restab)
}
##################################################

##################################################
#### run
restab <- main(inputFileName, geneFileName, exonFileName)
##################################################
